#include <iostream>
#include <fstream>
#include <sstream>
#include <GL/glew.h>
#include <SFML/Window.hpp>
#include "Matrix.hpp"

// Tout plein de variables globales trs moches, mais qui facilitent la vie pour les exemples simples

sf::Window window; // Contexte de rendu
Matrix4 worldViewProjectionMatrix; // Concatnation de la matrice world et projection
unsigned int shaderProgram; // Identifiant du programme de shader
unsigned int uniformID; // Emplacement de la variable uniform dans le shader
unsigned int vertexPositionAttrib; // Emplacement de l'attribut position dans le shader
unsigned int vertexColorAttrib; // Emplacement de l'attribut couleur dans le shader
unsigned int cubeBuffers[2]; // Les identifiants des buffer
unsigned int vertexArrayObject; // Indice du vertex array object

// Les trois premires valeurs de chaque ligne sont les couleurs et les trois suivantes les coordonnes
float CubeArray[48] = {
	1.0f, 0.0f, 0.0f, -1.0f, 1.0f, -1.0f,
	1.0f, 0.0f, 1.0f, -1.0f, -1.0f, -1.0f,
	1.0f, 1.0f, 1.0f, -1.0f, 1.0f, 1.0f,
	0.0f, 0.0f, 1.0f, -1.0f, -1.0f, 1.0f,
	1.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f,
	1.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f,
	1.0f, 0.0f, 1.0f, 1.0f, 1.0f, -1.0f,
	1.0f, 1.0f, 0.0f, 1.0f, -1.0f, -1.0f
};

// Les indices de vertices
unsigned int IndiceArray[36] = {
	0,1,2,2,1,3,
	4,5,6,6,5,7,
	3,1,5,5,1,7,
	0,2,6,6,2,4,
	6,7,0,0,7,1,
	2,3,4,4,3,5
};

std::string LoadSource (const std::string & fileName)
{
	std::ifstream fichier (fileName.c_str());

   if (fichier)
   {
      std::stringstream buffer;
      buffer << fichier.rdbuf();
      fichier.close();
		std::string source (buffer.str());
		source.push_back ('\0');
		return source;
	}
}

static void InitOpenGL ()
{
	// Initialisation de GLEW
	glewInit ();
	
	// On rgle le viewport
	glViewport (0, 0, 800, 600);

	// Initialisation de la SFML, qui va nous crer un contexte valide OpenGL3.x (attention, uniquement  partir de la version 2.0, disponible en bta
	//  l'heure o j'cris
	window.Create (sf::VideoMode (800, 600), "Vertex Array Objects avec OpenGL 3.x");

	// Cration des matrices de transformation. Pour plus de simplicit, les matrices ne bougeront jamais
	Matrix4 projectionMatrix;
	projectionMatrix.SetProjectionMatrix (60.0f, 1.0f, 1.0f, 500.0f);
	
	Matrix4 worldMatrix;
	worldMatrix.SetTransformationMatrix (0.0f, 0.0f, -5.0f);

	worldViewProjectionMatrix = projectionMatrix * worldMatrix;
}


static void LoadShaders ()
{
	// Cration des shaders
	unsigned int vertexShaderID = glCreateShader (GL_VERTEX_SHADER);
	unsigned int fragmentShaderID = glCreateShader (GL_FRAGMENT_SHADER);
	
	// On charge les sources
	std::string vertexShaderString = LoadSource ("vs.txt");
	std::string fragmentShaderString = LoadSource ("fs.txt");
	const char * vertexShader = vertexShaderString.c_str();
	const char * fragmentShader = fragmentShaderString.c_str();
	glShaderSource (vertexShaderID, 1, &vertexShader, NULL);
	glShaderSource (fragmentShaderID, 1, &fragmentShader, NULL);

	// On compile les shaders
	glCompileShader (vertexShaderID);
	glCompileShader (fragmentShaderID);
	
	// On vrifie qu'ils aient bien t compils
	GLint error;
	glGetShaderiv (vertexShaderID, GL_COMPILE_STATUS, &error);
	if (error != GL_TRUE)
		std::cout << "Probleme!";

	glGetShaderiv (fragmentShaderID, GL_COMPILE_STATUS, &error);
	if (error != GL_TRUE)
		std::cout << "Probleme!";
	
	// Cration du programme
	shaderProgram = glCreateProgram ();
	glAttachShader (shaderProgram, vertexShaderID);
	glAttachShader (shaderProgram, fragmentShaderID);

	// Linkage du programme
	glLinkProgram (shaderProgram);
	
	// On rcupre l'emplacement de la variable uniform dans le shader
	uniformID = glGetUniformLocation (shaderProgram, "ModelViewProjectionMatrix");
	
	// On rcupre les emplacements des deux attributs dans le shader
	vertexPositionAttrib = glGetAttribLocation (shaderProgram, "VertexPosition");
	vertexColorAttrib = glGetAttribLocation (shaderProgram, "VertexColor");

	/*
	Comme je l'ai prcis dans le tutoriel, il est possible de spcifier soi-mme ses emplacements. Ainsi, on aurait pu fixer, de manire arbitraire,
	que l'attribut VertexPosition serait attach  l'emplacement 1 et VertexColor  l'emplacement 2, ce qui nous aurait donn :

	vertexPositionAttrib = 1;
	vertexColorAttrib = 2;

	glBindAttribLocation (shaderProgram, vertexPositionAttrib, "VertexPosition");
	glBindAttribLocation (shaderProgram, vertexColorAttrib, "VertexColor");
	*/
}


static void CreateBuffers ()
{
	// Gnration des buffers
	glGenBuffers (2, cubeBuffers);

	// Buffer d'informations de vertices
	glBindBuffer(GL_ARRAY_BUFFER, cubeBuffers[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(CubeArray), CubeArray, GL_STATIC_DRAW);

	// Buffer d'indices
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cubeBuffers[1]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(IndiceArray), IndiceArray, GL_STATIC_DRAW);
}


static void CreateVAO ()
{
	// Cration du VAO
	glGenVertexArrays (1, &vertexArrayObject);

	// On bind le VAO
	glBindVertexArray (vertexArrayObject);

	// Et on appelle toutes les fonctions qui modifient l'tat des tableaux de sommets

	glEnableVertexAttribArray (vertexPositionAttrib);
	glEnableVertexAttribArray (vertexColorAttrib);
		
	// On active le buffer des vertices
	glBindBuffer(GL_ARRAY_BUFFER, cubeBuffers[0]);

	// On envoie les donnes
	glVertexAttribPointer (vertexColorAttrib, 3, GL_FLOAT, GL_FALSE, sizeof (float) * 6, (float *) NULL + (0));
	glVertexAttribPointer (vertexPositionAttrib, 3, GL_FLOAT, GL_FALSE, sizeof (float) * 6, (float *) NULL + (3));
		
	// On active le buffer des indices
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cubeBuffers[1]);

	// On unbind le VAO, tout ceci est enregistr
	glBindVertexArray (0);
}


static void Clean ()
{
	glDeleteProgram (shaderProgram);
	glDeleteBuffers (2, cubeBuffers);
	glDeleteVertexArrays (1, &vertexArrayObject);
}


int main ()
{
	// Initialisation d'OpenGL
	InitOpenGL ();

	// Chargement des shaders
	LoadShaders ();

	// Creation des buffers
	CreateBuffers ();

	// Cration du vertex array object
	CreateVAO ();

	// Boucle principale
   while (window.IsOpened())
   {
      // Process events
      sf::Event Event;
      while (window.GetEvent(Event))
      {
         // Close window : exit
         if (Event.Type == sf::Event::Closed)
            window.Close();

         // Escape key : exit
         if ((Event.Type == sf::Event::KeyPressed) && (Event.Key.Code == sf::Key::Escape))
            window.Close();
      }

		// On commence par activer le shader
		glUseProgram (shaderProgram);

		// On envoit la variable uniform
		glUniformMatrix4fv (uniformID, 1, GL_FALSE, worldViewProjectionMatrix.GetMatrixForOpenGL());

		/* TOUTES CES LIGNES DE CODE :
		
		// On active les deux tableaux gnriques
		glEnableVertexAttribArray (vertexPositionAttrib);
		glEnableVertexAttribArray (vertexColorAttrib);
		
		// On active le buffer des vertices
		glBindBuffer(GL_ARRAY_BUFFER, cubeBuffers[0]);

		// On envoie les donnes
		glVertexAttribPointer (vertexColorAttrib, 3, GL_FLOAT, GL_FALSE, sizeof (float) * 6, (float *) NULL + (0));
		glVertexAttribPointer (vertexPositionAttrib, 3, GL_FLOAT, GL_FALSE, sizeof (float) * 6, (float *) NULL + (3));
		
		// On active le buffer des indices
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cubeBuffers[1]);

		// Rendu de notre gomtrie
		glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_INT, 0);

		// On dsactive les tableaux gnriques
		glDisableVertexAttribArray (vertexPositionAttrib);
		glDisableVertexAttribArray (vertexColorAttrib);*/

		// SONT REMPLACEES PAR :
		glBindVertexArray (vertexArrayObject);

		glDrawElements (GL_TRIANGLES, 36, GL_UNSIGNED_INT, 0);
		
		// On peut ventuellement dsactvier les tableaux gnriques car on a pas enregistr ces commandes dans le VAO, mais ceci n'est
		// pas ncessaire pour une application 100% shader qui respecte OGL 3.x
		glBindVertexArray (0);

		// On dsactive le shader
		glUseProgram (0);


		/*
		NOTE : Normalement, dans un programme utilisant uniquement les shaders, tous ces "disable" et glUseProgram (0) ne devraient plus tre ncessaires,
		car on utilise exclusivement les shaders, l'activation d'un shader "annulera" de toute faon le prcdent qui a t activ
		*/


      // On affiche le tout !
      window.Display();
   }

	// On nettoie le tout
	Clean ();

	return EXIT_SUCCESS;
}